import logging

from ..helper.string import camel_case_to_snake_case
from ..helper.uid import get_device_uid_from_element_uid
from ..mydevolo import Mydevolo


class Zwave:
    """
    Representing object for Z-Wave devices. Basically, everything can be stored in here. This is to be as flexible to gateway
    firmware changes as possible. So if new attributes appear or old ones are removed, they should be handled at least in
    reading them. Nevertheless, a few unwanted attributes are filtered.

    :param mydevolo_instance: Mydevolo instance for talking to the devolo Cloud
    :key batteryLevel: Battery Level of the device in percent, -1 if mains powered
    :type batteryLevel: int
    :key elementUIDs: All element UIDs the device has
    :type elementUIDs: list
    :key manID: Manufacturer ID as registered at the Z-Wave alliance
    :type manID: str
    :key prodID: Product ID as registered at the Z-Wave alliance
    :type prodID: str
    :key prodTypeID: Product type ID as registered at the Z-Wave alliance
    :type prodTypeID: str
    :key status: Online status of the device
    :type status: int
    """

    def __init__(self, mydevolo_instance: Mydevolo, **kwargs):
        self._logger = logging.getLogger(self.__class__.__name__)
        self._mydevolo = mydevolo_instance

        # Get important values
        self.battery_level = kwargs.pop("batteryLevel", -1)
        self.element_uids = kwargs.pop("elementUIDs")
        self.man_id = kwargs.pop("manID")
        self.prod_id = kwargs.pop("prodID")
        self.prod_type_id = kwargs.pop("prodTypeID")
        self.status = kwargs.pop("status", 1)

        # Remove unwanted values
        unwanted_value = [
            "icon",
            "itemName",
            "operationStatus",
            "zone",
            "zoneId",
        ]
        for value in unwanted_value:
            del kwargs[value]

        # Add all other values
        for key, value in kwargs.items():
            setattr(self, camel_case_to_snake_case(key), value)
        self.uid = get_device_uid_from_element_uid(self.element_uids[0])

        # Initialize additional Z-Wave information. Will be filled by Zwave.get_zwave_info, if available.
        z_wave_info_list = [
            "href",
            "manufacturer_id",
            "product_type_id",
            "product_id",
            "name",
            "brand",
            "identifier",
            "is_zwave_plus",
            "device_type",
            "zwave_version",
            "specific_device_class",
            "generic_device_class",
        ]
        for key in z_wave_info_list:
            setattr(self, key, None)

        # Remove battery properties, if device is mains powered.
        if self.battery_level == -1:
            delattr(self, "battery_level")
            delattr(self, "battery_low")

    def get_property(self, name: str) -> list:
        """
        Get element UIDs to a specified property.

        :param name: Name of the property we want to access
        :return: List of UIDs in this property
        :raises: AttributeError: The property does not exist in this device type
        """
        return [*getattr(self, f"{name}_property").values()]

    def get_zwave_info(self):
        """
        Get publicly available information like manufacturer or model from my devolo. For a complete list, please look at
        Zwave.__init__.
        """
        self._logger.debug("Getting Z-Wave information for %s", self.uid)
        zwave_product = self._mydevolo.get_zwave_products(manufacturer=self.man_id,
                                                          product_type=self.prod_type_id,
                                                          product=self.prod_id)
        for key, value in zwave_product.items():
            setattr(self, camel_case_to_snake_case(key), value)

        # Clean up attributes which are now unwanted.
        clean_up_list = [
            "man_id",
            "prod_id",
            "prod_type_id",
            "statistics_uid",
            "wrong_device_paired",
        ]
        for attribute in clean_up_list:
            if hasattr(self, attribute):
                delattr(self, attribute)

    def is_online(self) -> bool:
        """
        Get the online state of a device.

        :return: False, if device is offline, else True
        """
        return self.status != 1



import pytest


@pytest.fixture()
def mock_socket_inet_ntoa(mocker, request):
    """ Mock socket's hostname to IP method to be independent from a real existing device. """
    mocker.patch("socket.inet_ntoa", return_value=request.cls.gateway.get("local_ip"))



import json
import pathlib

from .mock_dummy_device import dummy_device
from .mock_humidity_sensor_device import humidity_sensor_device
from .mock_metering_plug import metering_plug
from .mock_multi_level_sensor_device import multi_level_sensor_device
from .mock_multi_level_switch_device import multi_level_switch_device
from .mock_remote_control import remote_control
from .mock_shutter import shutter
from .mock_siren import siren


def mock__inspect_devices(self, devices):
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mapping = {
        "blinds": shutter,
        "humidity": humidity_sensor_device,
        "mains": metering_plug,
        "multi_level_switch": multi_level_switch_device,
        "remote": remote_control,
        "sensor": multi_level_sensor_device,
        "siren": siren,
    }

    for device_type, device in test_data['devices'].items():
        device_uid = device['uid']
        self.devices[device_uid] = mapping.get(device_type,
                                               dummy_device)(device_uid if device_type in mapping else device_type)



import shlex
from subprocess import check_call

from setuptools import find_packages, setup
from setuptools.command.develop import develop

with open("README.md", "r") as fh:
    long_description = fh.read()


# Create post develop command class for hooking into the python setup process
# This command will run after dependencies are installed
class PostDevelopCommand(develop):

    def run(self):
        try:
            check_call(shlex.split("pre-commit install"))
        except Exception:
            print("Unable to run 'pre-commit install'")
        develop.run(self)


setup(
    name="devolo_home_control_api",
    use_scm_version=True,
    author="Markus Bong, Guido Schmitz",
    author_email="m.bong@famabo.de, guido.schmitz@fedaix.de",
    description="devolo Home Control API in Python",
    license="GPLv3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/2Fake/devolo_home_control_api",
    packages=find_packages(exclude=("tests*",
                                    )),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "importlib-metadata;python_version<'3.8'",
        "requests",
        "websocket_client>=0.58.0",
        "zeroconf",
    ],
    setup_requires=["setuptools_scm"],
    extras_require={
        "dev": [
            "pre-commit",
        ],
        "test": [
            "pytest",
            "pytest-cov",
            "pytest-mock",
        ],
    },
    cmdclass={"develop": PostDevelopCommand},
    python_requires=">=3.6",
)



import pytest

from ..mocks.mock_gateway import MockGateway


@pytest.fixture()
def gateway_instance(request, mydevolo):
    """ Create a completly mocked gateway instance. """
    return MockGateway(request.cls.gateway.get("id"), mydevolo=mydevolo)


@pytest.fixture()
def mock_gateway(mocker):
    """ Mock only the constructor of a gateway instance. """
    mocker.patch("devolo_home_control_api.devices.gateway.Gateway.__init__", MockGateway.__init__)



import pytest
from devolo_home_control_api.exceptions.gateway import GatewayOfflineError


@pytest.mark.usefixtures("mprm_instance")
class TestMprmRest:

    @pytest.mark.usefixtures("mock_mprmrest__post")
    def test_get_name_and_element_uids(self):
        properties = self.mprm.get_name_and_element_uids("test")
        assert properties == {
            "itemName": "test_name",
            "zone": "test_zone",
            "batteryLevel": "test_battery",
            "icon": "test_icon",
            "elementUIDs": "test_element_uids",
            "settingUIDs": "test_setting_uids",
            "deviceModelUID": "test_device_model_uid",
            "status": "test_status",
        }

    @pytest.mark.usefixtures("mock_mprmrest__post")
    def test_get_all_devices(self):
        devices = self.mprm.get_all_devices()
        assert devices == "deviceUIDs"

    @pytest.mark.usefixtures("mock_mprmrest__post")
    def test_get_all_zones(self):
        zones = self.mprm.get_all_zones()
        assert zones == {
            "hz_3": "Office",
        }

    @pytest.mark.usefixtures("mock_response_requests_ReadTimeout")
    def test_post_ReadTimeOut(self, mprm_session, gateway_instance):
        self.mprm._session = mprm_session
        self.mprm.gateway = gateway_instance
        with pytest.raises(GatewayOfflineError):
            self.mprm._post({"data": "test"})

    @pytest.mark.usefixtures("mock_response_requests_invalid_id")
    def test_post_invalid_id(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._data_id = 0
        with pytest.raises(ValueError):
            self.mprm._post({"data": "test"})

    @pytest.mark.usefixtures("mock_response_requests_valid")
    def test_post_valid(self, mprm_session):
        self.mprm._session = mprm_session
        self.mprm._data_id = 1
        assert self.mprm._post({
            "data": "test"
        }).get("id") == 2

    @pytest.mark.usefixtures("mock_mprmrest__post")
    @pytest.mark.parametrize("setter",
                             [
                                 ("set_binary_switch"),
                                 ("set_multi_level_switch"),
                                 ("set_remote_control"),
                                 ("set_setting"),
                             ])
    def test_set_success(self, setter):
        test_data = {
            "set_binary_switch": bool(self.devices['mains']['properties']['state']),
            "set_multi_level_switch": self.devices['multi_level_switch']['value'],
            "set_remote_control": 1,
            "set_setting": self.devices['mains']['properties']['local_switch'],
        }
        assert getattr(self.mprm, setter)(self.devices['mains']['properties']['elementUIDs'][1], test_data[setter])

    @pytest.mark.usefixtures("mock_mprmrest__post")
    @pytest.mark.parametrize("setter", [("set_binary_switch"), ("set_multi_level_switch")])
    def test_set_doubled(self, setter):
        assert not getattr(self.mprm,
                           setter)(self.devices['mains']['properties']['elementUIDs'][1],
                                   bool(self.devices['mains']['properties']['state']))

    @pytest.mark.usefixtures("mock_mprmrest__post")
    @pytest.mark.parametrize("setter",
                             [
                                 ("set_binary_switch"),
                                 ("set_multi_level_switch"),
                                 ("set_remote_control"),
                                 ("set_setting"),
                             ])
    def test_set_failed(self, setter):
        assert not getattr(self.mprm,
                           setter)(self.devices['mains']['properties']['elementUIDs'][1],
                                   bool(self.devices['mains']['properties']['state']))

    @pytest.mark.usefixtures("mock_mprmrest__post")
    def test_get_data_from_uid_list(self):
        properties = self.mprm.get_data_from_uid_list(["test"])
        assert properties[0].get("properties").get("itemName") == "test_name"



import json
import pathlib

from devolo_home_control_api.devices.zwave import Zwave
from devolo_home_control_api.mydevolo import Mydevolo
from devolo_home_control_api.properties.remote_control_property import RemoteControlProperty
from devolo_home_control_api.properties.settings_property import SettingsProperty


def remote_control(device_uid: str) -> Zwave:
    """
    Represent a remote control in tests.

    :param device_uid: Device UID this mock shall have
    :return: Remote Control
    """
    file = pathlib.Path(__file__).parent / ".." / "test_data.json"
    with file.open("r") as fh:
        test_data = json.load(fh)

    mydevolo = Mydevolo()
    device = Zwave(mydevolo_instance=mydevolo, **test_data['devices']['remote'])

    device.remote_control_property = {}
    device.settings_property = {}

    device.remote_control_property[f'devolo.RemoteControl:{device_uid}'] = RemoteControlProperty(
        element_uid=f"devolo.RemoteControl:{device_uid}",
        setter=lambda uid,
        state: True,
        key_count=test_data['devices']['remote']['key_count'],
        key_pressed=0)

    device.settings_property["general_device_settings"] = SettingsProperty(element_uid=f"gds.{device_uid}",
                                                                           setter=lambda uid,
                                                                           state: None,
                                                                           icon=test_data['devices']['remote']['icon'],
                                                                           name=test_data['devices']['remote']['itemName'],
                                                                           zone_id=test_data['devices']['remote']['zoneId'],
                                                                           zones=test_data['gateway']['zones'])

    device.settings_property["switch_type"] = SettingsProperty(element_uid=f"sts.{device_uid}",
                                                               setter=lambda uid,
                                                               state: None,
                                                               value=test_data['devices']['remote']['key_count'])

    return device



from typing import Callable

from ..exceptions.device import WrongElementError
from .property import Property


class SettingsProperty(Property):
    """
    Object for settings. Basically, everything can be stored in here as long as there is a corresponding functional item on
    the gateway. This is to be as flexible to gateway firmware changes as possible. So if new attributes appear or old ones
    are removed, they should be handled at least in reading them. Nevertheless, a few unwanted attributes are filtered.

    :param element_uid: Element UID, something like acs.hdm:ZWave:CBC56091/24
    :param setter: Method to call on setting the state
    :param **kwargs: Any setting, that shall be available in this object
    """

    def __init__(self, element_uid: str, setter: Callable, **kwargs):
        if not element_uid.startswith(("acs",
                                       "bas",
                                       "bss",
                                       "cps",
                                       "gds",
                                       "lis",
                                       "mas",
                                       "mss",
                                       "ps",
                                       "sts",
                                       "stmss",
                                       "trs",
                                       "vfs")):
            raise WrongElementError()

        super().__init__(element_uid=element_uid)
        self._setter = setter

        if element_uid.startswith("gds") and {"zones",
                                              "zone_id"} <= kwargs.keys():
            self.events_enabled: bool
            self.icon: str
            self.name: str
            self.zone_id: str
            self.zone = kwargs.pop("zones")[kwargs['zone_id']]

        for key, value in kwargs.items():
            setattr(self, key, value)

        setter_method = {
            "bas": self._set_bas,
            "gds": self._set_gds,
            "lis": self._set_lis,
            "mss": self._set_mss,
            "ps": self._set_ps,
            "trs": self._set_trs,
            "vfs": self._set_lis,
        }

        # Depending on the type of setting property, this will create a callable named "set".
        # However, this methods are not working, if the gateway is connected locally, yet.
        self.set = setter_method.get(element_uid.split(".")[0])

        # Clean up attributes which are unwanted.
        clean_up_list = ["device_uid"]

        for attribute in clean_up_list:
            delattr(self, attribute)

    def _set_bas(self, value: bool):
        """
        Set a binary async setting. This is e.g. the muted setting of a siren or the three way switch setting of a dimmer.

        :param value: New state
        """
        if self._setter(self.element_uid, [value]):
            self.value = value  # pylint: disable=attribute-defined-outside-init
            self._logger.debug("Binary async setting property %s set to %s", self.element_uid, value)

    def _set_gds(self, **kwargs):
        """
        Set one or more general device setting.

        :key events_enabled: Show events in diary
        :type events_enabled: bool
        :key icon: New icon name
        :type icon: str
        :key name: New device name
        :type name: str
        :key zone_id: New zone_id (ATTENTION: This is NOT the name of the location)
        :type zone_id: str
        """
        events_enabled: bool = kwargs.pop("events_enabled", self.events_enabled)
        icon: str = kwargs.pop("icon", self.icon)
        name: str = kwargs.pop("name", self.name)
        zone_id: str = kwargs.pop("zone_id", self.zone_id)

        settings = {
            "events_enabled": events_enabled,
            "icon": icon,
            "name": name,
            "zone_id": zone_id,
        }
        if self._setter(self.element_uid, [settings]):
            self.events_enabled = events_enabled
            self.icon = icon
            self.name = name
            self.zone_id = zone_id
            self._logger.debug("General device setting %s changed.", self.element_uid)

    def _set_lis(self, led_setting: bool):
        """
        Set led settings.

        :param led_setting: LED indication setting
        """
        if self._setter(self.element_uid, [led_setting]):
            self.led_setting = led_setting  # pylint: disable=attribute-defined-outside-init
            self._logger.debug("LED indication setting property %s set to %s", self.element_uid, led_setting)

    def _set_mss(self, motion_sensitivity: int):
        """
        Set motion sensitivity.

        :param motion_sensitivity: Integer for the motion sensitivity setting.
        """
        if not 0 <= motion_sensitivity <= 100:
            raise ValueError("Value must be between 0 and 100")
        if self._setter(self.element_uid, [motion_sensitivity]):
            self.motion_sensitivity = motion_sensitivity  # pylint: disable=attribute-defined-outside-init
            self._logger.debug("Motion sensitivity setting property %s set to %s", self.element_uid, motion_sensitivity)

    def _set_ps(self, **kwargs: bool):
        """
        Set one or both protection settings.

        :key local_switching: Allow local switching
        :type local_switching: bool
        :key remote_switching: Allow local switching
        :type remote_switching: bool
        """
        # pylint: disable=access-member-before-definition
        local_switching = kwargs.pop("local_switching", self.local_switching)
        # pylint: disable=access-member-before-definition
        remote_switching = kwargs.pop("remote_switching", self.remote_switching)

        if self._setter(self.element_uid,
                        [{
                            "localSwitch": local_switching,
                            "remoteSwitch": remote_switching,
                        }]):
            self.local_switching: bool = local_switching  # pylint: disable=attribute-defined-outside-init
            self.remote_switching: bool = remote_switching  # pylint: disable=attribute-defined-outside-init
            self._logger.debug("Protection setting property %s set to %s (local) and %s (remote).",
                               self.element_uid,
                               local_switching,
                               remote_switching)

    def _set_trs(self, temp_report: bool):
        """
        Set temperature report setting.

        :param temp_report: Boolean of the target value
        """
        if self._setter(self.element_uid, [temp_report]):
            self.temp_report = temp_report  # pylint: disable=attribute-defined-outside-init
            self._logger.debug("Temperature report setting property %s set to %s", self.element_uid, temp_report)



from datetime import datetime

from ..exceptions.device import WrongElementError
from .sensor_property import SensorProperty


class BinarySensorProperty(SensorProperty):
    """
    Object for binary sensors. It stores the binary sensor state.

    :param element_uid: Element UID, something like devolo.BinarySensor:hdm:ZWave:CBC56091/24
    :key state: State of the binary sensor
    :type state: bool
    """

    def __init__(self, element_uid: str, **kwargs):
        if not element_uid.startswith(("devolo.BinarySensor:",
                                       "devolo.MildewSensor:",
                                       "devolo.ShutterMovementFI:",
                                       "devolo.WarningBinaryFI:")):
            raise WrongElementError(f"{element_uid} is not a Binary Sensor.")

        super().__init__(element_uid=element_uid, **kwargs)

        self._state: bool = kwargs.pop("state", False)

    @property
    def last_activity(self) -> datetime:
        """ Date and time the state of the binary sensor was last updated. """
        return super().last_activity

    @last_activity.setter
    def last_activity(self, timestamp: int):
        """ The gateway persists the last activity of some binary sensors. They can be initialized with that value. """
        if timestamp != -1:
            self._last_activity = datetime.utcfromtimestamp(timestamp / 1000)
            self._logger.debug("last_activity of element_uid %s set to %s.", self.element_uid, self._last_activity)

    @property
    def state(self) -> bool:
        """ State of the binary sensor. """
        return self._state

    @state.setter
    def state(self, state: bool):
        """ Update state of the binary sensor and set point in time of the last_activity. """
        self._state = state
        self._last_activity = datetime.now()
        self._logger.debug("state of element_uid %s set to %s.", self.element_uid, state)



import json
import pathlib

file = pathlib.Path(__file__).parent / ".." / "test_data.json"
with file.open("r") as fh:
    test_data = json.load(fh)


def try_local_connection(self, mdns_name):
    self._local_ip = test_data["gateway"]["local_ip"]


def mock_get_data_from_uid_list(self, uids):
    if len(uids) == 1:
        return [test_data["devices"]["mains"]]
    else:
        return [{
            "UID": test_data["devices"]["mains"]["settingUIDs"][0],
            "properties": {
                "settings": test_data["devices"]["mains"]["properties"]
            },
        }]



class MockMydevolo:

    def __init__(self, request):
        self._request = request

    def _call(self, url):
        uuid = self._request.cls.user.get("uuid")
        gateway_id = self._request.cls.gateway.get("id")
        full_url = self._request.cls.gateway.get("full_url")

        response = {
            f"https://www.mydevolo.com/v1/users/{uuid}/hc/gateways/{gateway_id}/fullURL": {
                "url": full_url
            },
            "https://www.mydevolo.com/v1/users/uuid": {
                "uuid": uuid
            },
            f"https://www.mydevolo.com/v1/users/{uuid}/hc/gateways/status": {
                "items": []
            } if self._request.node.name == "test_gateway_ids_empty" else {
                "items": [{
                    "gatewayId": gateway_id
                }]
            },
            f"https://www.mydevolo.com/v1/users/{uuid}/hc/gateways/{gateway_id}": {
                "gatewayId": gateway_id,
                "localPasskey": "abcde",
                "status": "devolo.hc_gateway.status.online",
                "state": "devolo.hc_gateway.state.idle",
            },
            "https://www.mydevolo.com/v1/hc/maintenance": {
                "state": "off"
            } if self._request.node.name == "test_maintenance[True]" else {
                "state": "on"
            },
            "https://www.mydevolo.com/v1/zwave/products/0x0060/0x0001/0x000": {
                "brand": "Everspring",
                "deviceType": "Door Lock Keypad",
                "genericDeviceClass": "Entry Control",
                "identifier": "SP814-US",
                "isZWavePlus": True,
                "manufacturerId": "0x0060",
                "name": "Everspring PIR Sensor SP814",
                "productId": "0x0002",
                "productTypeId": "0x0001",
                "zwaveVersion": "6.51.07",
            },
            "https://www.mydevolo.com/v1/zwave/products/0x0175/0x0001/0x0011": {
                "manufacturerId": "0x0175",
                "productTypeId": "0x0001",
                "productId": "0x0011",
                "name": "Metering Plug",
                "brand": "devolo",
                "identifier": "MT02646",
                "isZWavePlus": True,
                "deviceType": "On/Off Power Switch",
                "zwaveVersion": "6.51.00",
                "specificDeviceClass": None,
                "genericDeviceClass": None,
            },
        }

        return response[url]



import pytest
from devolo_home_control_api.exceptions.device import WrongElementError
from devolo_home_control_api.properties.consumption_property import ConsumptionProperty


@pytest.mark.usefixtures("home_control_instance")
class TestConsumption:

    def test_consumption_property_invalid(self):
        with pytest.raises(WrongElementError):
            ConsumptionProperty(element_uid="invalid", setter=lambda uid, state: None)



import threading
from typing import Dict, Optional

import requests
from zeroconf import Zeroconf

from . import __version__
from .backend import MESSAGE_TYPES
from .backend.mprm import Mprm
from .devices.gateway import Gateway
from .devices.zwave import Zwave
from .helper.string import camel_case_to_snake_case
from .helper.uid import (get_device_type_from_element_uid,
                         get_device_uid_from_element_uid,
                         get_device_uid_from_setting_uid,
                         get_home_id_from_device_uid)
from .mydevolo import Mydevolo
from .properties.binary_sensor_property import BinarySensorProperty
from .properties.binary_switch_property import BinarySwitchProperty
from .properties.consumption_property import ConsumptionProperty
from .properties.humidity_bar_property import HumidityBarProperty
from .properties.multi_level_sensor_property import MultiLevelSensorProperty
from .properties.multi_level_switch_property import MultiLevelSwitchProperty
from .properties.remote_control_property import RemoteControlProperty
from .properties.settings_property import SettingsProperty
from .publisher.publisher import Publisher
from .publisher.updater import Updater


class HomeControl(Mprm):
    """
    Representing object for your Home Control setup. This is more or less the glue between your devolo Home Control Central
    Unit, your devices and their properties.

    :param gateway_id: Gateway ID (aka serial number), typically found on the label of the device
    :param mydevolo_instance: Mydevolo instance for talking to the devolo Cloud
    :param zeroconf_instance: Zeroconf instance to be potentially reused
    """

    def __init__(self, gateway_id: str, mydevolo_instance: Mydevolo, zeroconf_instance: Optional[Zeroconf] = None):
        self._mydevolo = mydevolo_instance
        self._session = requests.Session()
        self._session.headers.update({"User-Agent": f"devolo_home_control_api/{__version__}"})
        self._zeroconf = zeroconf_instance
        self.gateway = Gateway(gateway_id, mydevolo_instance)

        super().__init__()
        self._grouping()

        # Create the initial device dict
        self.devices: Dict = {}
        self._inspect_devices(self.get_all_devices())

        self.device_names = {
            f"{self.devices[device].settings_property['general_device_settings'].name}/\
              {self.devices[device].settings_property['general_device_settings'].zone}": self.devices[device].uid
            for device in self.devices
        }

        self.gateway.home_id = get_home_id_from_device_uid(next(iter(self.device_names.values())))

        self.publisher = Publisher(self.devices.keys())

        self.updater = Updater(devices=self.devices, gateway=self.gateway, publisher=self.publisher)
        self.updater.on_device_change = self.device_change

        threading.Thread(target=self.websocket_connect, name=f"{self.__class__.__name__}.websocket_connect").start()
        self.wait_for_websocket_establishment()

    @property
    def binary_sensor_devices(self) -> list:
        """ Get all binary sensor devices. """
        return [self.devices[uid] for uid in self.devices if hasattr(self.devices[uid], "binary_sensor_property")]

    @property
    def binary_switch_devices(self) -> list:
        """ Get all binary switch devices. """
        return [self.devices[uid] for uid in self.devices if hasattr(self.devices[uid], "binary_switch_property")]

    @property
    def blinds_devices(self) -> list:
        """ Get all blinds devices. """
        blinds_devices = []
        for device in self.multi_level_switch_devices:
            blinds_devices.extend([
                self.devices[device.uid]
                for multi_level_switch_property in device.multi_level_switch_property
                if multi_level_switch_property.startswith("devolo.Blinds")
            ])
        return blinds_devices

    @property
    def multi_level_sensor_devices(self) -> list:
        """ Get all multi level sensor devices. """
        return [self.devices[uid] for uid in self.devices if hasattr(self.devices[uid], "multi_level_sensor_property")]

    @property
    def multi_level_switch_devices(self) -> list:
        """ Get all multi level switch devices. This also includes blinds devices. """
        return [self.devices[uid] for uid in self.devices if hasattr(self.devices[uid], "multi_level_switch_property")]

    @property
    def remote_control_devices(self) -> list:
        """ Get all remote control devices. """
        return [self.devices[uid] for uid in self.devices if hasattr(self.devices[uid], "remote_control_property")]

    def device_change(self, device_uids: list):
        """
        React on new devices or removed devices. As the Z-Wave controller can only be in inclusion or exclusion mode, we
        assume, that you cannot add and remove devices at the same time. So if the number of devices increases, there is
        a new one and if the number decreases, a device was removed.

        :param device_uids: List of UIDs known by the backend
        """
        if len(device_uids) > len(self.devices):
            devices = [device for device in device_uids if device not in self.devices]
            mode = "add"
            self._inspect_devices([devices[0]])
            self._logger.debug("Device %s added.", devices[0])
        else:
            devices = [device for device in self.devices if device not in device_uids]
            mode = "del"
            self.devices.pop(devices[0])
            self._logger.debug("Device %s removed.", devices[0])
        self.updater.devices = self.devices
        return (devices[0], mode)

    def on_update(self, message):
        """
        Initialize steps needed to update properties on a new message.

        :param message: Message because of which we need to update properties
        """
        self.updater.update(message)

    def _binary_sensor(self, uid_info: dict):
        """ Process BinarySensor properties"""
        device_uid = get_device_uid_from_element_uid(uid_info['UID'])
        if not hasattr(self.devices[device_uid], "binary_sensor_property"):
            self.devices[device_uid].binary_sensor_property = {}
        self._logger.debug("Adding binary sensor property to %s.", device_uid)
        self.devices[device_uid].binary_sensor_property[uid_info['UID']] = BinarySensorProperty(
            element_uid=uid_info['UID'],
            state=bool(uid_info['properties']['state']),
            sensor_type=uid_info['properties']['sensorType'],
            sub_type=uid_info['properties']['subType'])

    def _binary_switch(self, uid_info: dict):
        """ Process BinarySwitch properties. """
        device_uid = get_device_uid_from_element_uid(uid_info['UID'])
        if not hasattr(self.devices[device_uid], "binary_switch_property"):
            self.devices[device_uid].binary_switch_property = {}
        self._logger.debug("Adding binary switch property to %s.", device_uid)
        self.devices[device_uid].binary_switch_property[uid_info['UID']] = BinarySwitchProperty(
            element_uid=uid_info['UID'],
            setter=self.set_binary_switch,
            state=bool(uid_info['properties']['state']),
            enabled=uid_info['properties']['guiEnabled'])

    def _general_device(self, uid_info: dict):
        """ Process general device setting (gds) properties. """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding general device settings to %s.", device_uid)
        self.devices[device_uid].settings_property['general_device_settings'] = SettingsProperty(
            element_uid=uid_info['UID'],
            setter=self.set_setting,
            events_enabled=uid_info['properties']['settings']['eventsEnabled'],
            name=uid_info['properties']['settings']['name'],
            zone_id=uid_info['properties']['settings']['zoneID'],
            icon=uid_info['properties']['settings']['icon'],
            zones=self.gateway.zones)

    def _grouping(self):
        """ Get all zones (also called rooms). """
        self.gateway.zones = self.get_all_zones()

    def _humidity_bar(self, uid_info: dict):
        """
        Process HumidityBarZone and HumidityBarValue properties.

        For whatever reason, the zone and the position within that zone are two different FIs. Nevertheless, it does not make
        sense to separate them. So we fake an FI and a sensorType to combine them together into one object.
        """
        device_uid = get_device_uid_from_element_uid(uid_info['UID'])
        fake_element_uid = f"devolo.HumidityBar:{uid_info['UID'].split(':', 1)[1]}"
        if not hasattr(self.devices[device_uid], "humidity_bar_property"):
            self.devices[device_uid].humidity_bar_property = {}
        if self.devices[device_uid].humidity_bar_property.get(fake_element_uid) is None:
            self.devices[device_uid].humidity_bar_property[fake_element_uid] = HumidityBarProperty(
                element_uid=fake_element_uid,
                sensorType="humidityBar")
        if uid_info['properties']['sensorType'] == "humidityBarZone":
            self._logger.debug("Adding humidity bar zone property to %s.", device_uid)
            self.devices[device_uid].humidity_bar_property[fake_element_uid].zone = uid_info['properties']['value']
        elif uid_info['properties']['sensorType'] == "humidityBarPos":
            self._logger.debug("Adding humidity bar position property to %s.", device_uid)
            self.devices[device_uid].humidity_bar_property[fake_element_uid].value = uid_info['properties']['value']

    def _inspect_devices(self, devices: list):
        """ Inspect device properties of given list of devices. """
        devices_properties = self.get_data_from_uid_list(devices)
        for device_properties in devices_properties:
            properties = device_properties['properties']
            self.devices[device_properties['UID']] = Zwave(mydevolo_instance=self._mydevolo, **properties)
            self.devices[device_properties['UID']].settings_property = {}
            threading.Thread(target=self.devices[device_properties['UID']].get_zwave_info,
                             name=f"{self.__class__.__name__}.{self.devices[device_properties['UID']].uid}").start()

        # List comprehension gets the list of uids from every device
        nested_uids_lists = [(uid['properties'].get("settingUIDs") + uid['properties']['elementUIDs'])
                             for uid in devices_properties]

        # List comprehension gets all uids into one list to make one big call against the mPRM
        uid_list = [uid for sublist in nested_uids_lists for uid in sublist]

        device_properties_list = self.get_data_from_uid_list(uid_list)

        for uid_info in device_properties_list:
            message_type = MESSAGE_TYPES.get(get_device_type_from_element_uid(uid_info['UID']), "_unknown")
            getattr(self, message_type)(uid_info)
            try:
                uid = self.devices[get_device_uid_from_element_uid(uid_info['UID'])]
            except KeyError:
                uid = self.devices[get_device_uid_from_setting_uid(uid_info['UID'])]
            uid.pending_operations = uid.pending_operations or bool(uid_info['properties'].get("pendingOperations"))

        # Last activity messages sometimes arrive before a device was initialized and therefore need to be handled afterwards.
        [  # pylint: disable=expression-not-assigned
            self._last_activity(uid_info)
            for uid_info in device_properties_list
            if uid_info['UID'].startswith("devolo.LastActivity")
        ]

    def _automatic_calibration(self, uid_info: dict):
        """ Process automatic calibration (acs) properties. """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding automatic calibration setting to %s.", device_uid)
        self.devices[device_uid].settings_property['automatic_calibration'] = SettingsProperty(
            element_uid=uid_info['UID'],
            setter=self.set_setting,
            calibration_status=bool(uid_info['properties']['calibrationStatus']))

    def _binary_sync(self, uid_info: dict):
        """
        Process binary sync setting (bss) properties. Currently only the direction of a shutter in known to be a binary sync
        setting property, so it is named nicely.
        """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding binary sync setting to %s.", device_uid)
        self.devices[device_uid].settings_property['movement_direction'] = SettingsProperty(
            element_uid=uid_info['UID'],
            setter=self.set_setting,
            inverted=uid_info['properties']['value'])

    def _binary_async(self, uid_info: dict):
        """ Process binary async setting (bas) properties. """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding binary async settings to %s.", device_uid)
        settings_property = SettingsProperty(element_uid=uid_info['UID'],
                                             setter=self.set_setting,
                                             value=uid_info['properties']['value'])

        # The siren needs to be handled differently, as otherwise their binary async setting will not be named nicely
        if self.devices[device_uid].device_model_uid == "devolo.model.Siren":
            self.devices[device_uid].settings_property['muted'] = settings_property
        # As some devices have multiple binary async settings, we use the settings UID split after a '#' as key
        else:
            key = camel_case_to_snake_case(uid_info['UID'].split("#")[-1])
            self.devices[device_uid].settings_property[key] = settings_property

    def _last_activity(self, uid_info: dict):
        """
        Process last activity properties. Those don't go into an own property but will be appended to a parent property.
        This parent property is found by string replacement.
        """
        device_uid = get_device_uid_from_element_uid(uid_info['UID'])
        parent_element_uid = uid_info['UID'].replace("LastActivity", "BinarySensor")
        try:
            self.devices[device_uid].binary_sensor_property[parent_element_uid].last_activity = \
                uid_info['properties']['lastActivityTime']
        except AttributeError:
            parent_element_uid = uid_info['UID'].replace("LastActivity", "SirenMultiLevelSwitch")
            self.devices[device_uid].multi_level_switch_property[parent_element_uid].last_activity = \
                uid_info['properties']['lastActivityTime']

    def _led(self, uid_info: dict):
        """ Process LED information setting (lis) and visual feedback setting (vfs) properties. """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding led settings to %s.", device_uid)
        try:
            led_setting = uid_info['properties']['led']
        except KeyError:
            led_setting = uid_info['properties']['feedback']
        self.devices[device_uid].settings_property['led'] = SettingsProperty(element_uid=uid_info['UID'],
                                                                             setter=self.set_setting,
                                                                             led_setting=led_setting)

    def _meter(self, uid_info: dict):
        """ Process meter properties. """
        device_uid = get_device_uid_from_element_uid(uid_info['UID'])
        if not hasattr(self.devices[device_uid], "consumption_property"):
            self.devices[device_uid].consumption_property = {}
        self._logger.debug("Adding consumption property to %s.", device_uid)
        self.devices[device_uid].consumption_property[uid_info['UID']] = ConsumptionProperty(
            element_uid=uid_info['UID'],
            current=uid_info['properties']['currentValue'],
            total=uid_info['properties']['totalValue'],
            total_since=uid_info['properties']['sinceTime'])

    def _multilevel_async(self, uid_info: dict):
        """ Process multilevel async setting (mas) properties. """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        try:
            name = camel_case_to_snake_case(uid_info['properties']['itemId'])
        # The Metering Plug has an multilevel async setting without an ID
        except TypeError:
            if self.devices[device_uid].device_model_uid == "devolo.model.Wall:Plug:Switch:and:Meter":
                name = "flash_mode"
            else:
                raise

        self._logger.debug("Adding %s setting to %s.", name, device_uid)
        self.devices[device_uid].settings_property[name] = SettingsProperty(element_uid=uid_info['UID'],
                                                                            setter=self.set_setting,
                                                                            value=uid_info['properties']['value'])

    def _multilevel_sync(self, uid_info: dict):
        """ Process multilevel sync setting (mss) properties. """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])

        # The siren needs to be handled differently, as otherwise their multilevel sync setting will not be named nicely.
        if self.devices[device_uid].device_model_uid == "devolo.model.Siren":
            self._logger.debug("Adding tone settings to %s.", device_uid)
            self.devices[device_uid].settings_property['tone'] = SettingsProperty(element_uid=uid_info['UID'],
                                                                                  setter=self.set_setting,
                                                                                  tone=uid_info['properties']['value'])

        # The shutter needs to be handled differently, as otherwise their multilevel sync setting will not be named nicely.
        elif self.devices[device_uid].device_model_uid in ("devolo.model.OldShutter", "devolo.model.Shutter"):
            self._logger.debug("Adding shutter duration settings to %s.", device_uid)
            self.devices[device_uid].settings_property['shutter_duration'] = SettingsProperty(
                element_uid=uid_info['UID'],
                setter=self.set_setting,
                shutter_duration=uid_info['properties']['value'])

        # Other devices are up to now always motion sensors.
        else:
            self._logger.debug("Adding motion sensitivity settings to %s.", device_uid)
            self.devices[device_uid].settings_property['motion_sensitivity'] = SettingsProperty(
                element_uid=uid_info['UID'],
                setter=self.set_setting,
                motion_sensitivity=uid_info['properties']['value'])

    def _multi_level_sensor(self, uid_info: dict):
        """ Process multi level sensor properties. """
        device_uid = get_device_uid_from_element_uid(uid_info['UID'])
        if not hasattr(self.devices[device_uid], "multi_level_sensor_property"):
            self.devices[device_uid].multi_level_sensor_property = {}
        self._logger.debug("Adding multi level sensor property %s to %s.", uid_info['UID'], device_uid)
        self.devices[device_uid].multi_level_sensor_property[uid_info['UID']] = MultiLevelSensorProperty(
            element_uid=uid_info['UID'],
            value=uid_info['properties']['value'],
            unit=uid_info['properties']['unit'],
            sensor_type=uid_info['properties']['sensorType'])

    def _multi_level_switch(self, uid_info: dict):
        """ Process multi level switch properties. """
        device_uid = get_device_uid_from_element_uid(uid_info['UID'])
        if not hasattr(self.devices[device_uid], "multi_level_switch_property"):
            self.devices[device_uid].multi_level_switch_property = {}
        self._logger.debug("Adding multi level switch property %s to %s.", uid_info['UID'], device_uid)
        self.devices[device_uid].multi_level_switch_property[uid_info['UID']] = MultiLevelSwitchProperty(
            element_uid=uid_info['UID'],
            setter=self.set_multi_level_switch,
            value=uid_info['properties']['value'],
            switch_type=uid_info['properties']['switchType'],
            max=uid_info['properties']['max'],
            min=uid_info['properties']['min'])

    def _parameter(self, uid_info: dict):
        """ Process custom parameter setting (cps) properties."""
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding parameter settings to %s.", device_uid)
        self.devices[device_uid].settings_property['param_changed'] = SettingsProperty(
            element_uid=uid_info['UID'],
            setter=self.set_setting,
            param_changed=uid_info['properties']['paramChanged'])

    def _protection(self, uid_info: dict):
        """ Process protection setting (ps) properties. """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding protection settings to %s.", device_uid)
        self.devices[device_uid].settings_property['protection'] = SettingsProperty(
            element_uid=uid_info['UID'],
            setter=self.set_setting,
            local_switching=uid_info['properties']['localSwitch'],
            remote_switching=uid_info['properties']['remoteSwitch'])

    def _remote_control(self, uid_info: dict):
        """ Process remote control properties. """
        device_uid = get_device_uid_from_element_uid(uid_info['UID'])
        self._logger.debug("Adding remote control to %s.", device_uid)
        if not hasattr(self.devices[device_uid], "remote_control_property"):
            self.devices[device_uid].remote_control_property = {}
        self.devices[device_uid].remote_control_property[uid_info['UID']] = RemoteControlProperty(
            element_uid=uid_info['UID'],
            setter=self.set_remote_control,
            key_count=uid_info['properties']['keyCount'],
            key_pressed=uid_info['properties']['keyPressed'],
            type=uid_info['properties']['type'])

    def _switch_type(self, uid_info: dict):
        """
        Process switch type setting (sts) properties. Interestingly, a switch with two buttons reports a switchType of 1 and a
        switch with four buttons reports a switchType of 2. This confusing behavior is corrected by doubling the value.
        """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding switch type setting to %s.", device_uid)
        self.devices[device_uid].settings_property['switch_type'] = SettingsProperty(element_uid=uid_info['UID'],
                                                                                     setter=self.set_setting,
                                                                                     value=uid_info['properties']['switchType']
                                                                                     * 2)

    def _temperature_report(self, uid_info: dict):
        """ Process temperature report setting (trs) properties. """
        device_uid = get_device_uid_from_setting_uid(uid_info['UID'])
        self._logger.debug("Adding temperature report settings to %s.", device_uid)
        self.devices[device_uid].settings_property['temperature_report'] = SettingsProperty(
            element_uid=uid_info['UID'],
            setter=self.set_setting,
            temp_report=uid_info['properties']['tempReport'],
            target_temp_report=uid_info['properties']['targetTempReport'])

    def _unknown(self, uid_info: dict):
        """ Ignore unknown properties. """
        ignore = ("devolo.SirenBinarySensor", "devolo.SirenMultiLevelSensor", "ss", "mcs")
        if not uid_info['UID'].startswith(ignore):
            self._logger.debug("Found an unexpected element uid: %s", uid_info['UID'])



from datetime import datetime, timezone

import pytest
from devolo_home_control_api.backend import MESSAGE_TYPES


@pytest.mark.usefixtures("home_control_instance")
@pytest.mark.usefixtures("mock_publisher_dispatch")
class TestUpdater:

    def test_hasattr(self):
        updater_functions = (func for func in dir(self.homecontrol.updater)
                             if callable(getattr(self.homecontrol.updater,
                                                 func)) and not func.startswith("__"))
        assert set(MESSAGE_TYPES.values()).difference(updater_functions) == set()

    @pytest.mark.usefixtures("mock_updater_binary_switch")
    def test_update_device(self, mocker):
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED",
            "properties": {
                "property.name": "state",
                "uid": f"devolo.BinarySwitch:{self.devices['mains']['uid']}",
            },
        }
        spy = mocker.spy(self.homecontrol.updater, '_binary_switch')
        self.homecontrol.updater.update(message=message)
        spy.assert_called_once_with(message)

    @pytest.mark.usefixtures("mock_updater_pending_operations")
    def test_update_pending_operations(self, mocker):
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED",
            "properties": {
                "property.name": "pendingOperations",
                "uid": "",
            },
        }
        spy = mocker.spy(self.homecontrol.updater, '_pending_operations')
        self.homecontrol.updater.update(message=message)
        spy.assert_called_once_with(message)

    def test_update_unwanted(self, mocker):
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/UNREGISTERED",
        }
        spy = mocker.spy(self.homecontrol.updater, '_unknown')
        self.homecontrol.updater.update(message=message)
        spy.assert_not_called()

    def test_update_smart_group(self, mocker):
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED",
            "properties": {
                "uid": "devolo.MultiLevelSwitch:devolo.smartGroup.Group",
                "property.name": "value",
            },
        }
        spy = mocker.spy(self.homecontrol.updater, '_multi_level_switch')
        self.homecontrol.updater.update(message=message)
        spy.assert_not_called()
        message = {
            "topic": "com/prosyst/mbs/services/fim/FunctionalItemEvent/PROPERTY_CHANGED",
            "properties": {
                "uid": "devolo.BinarySwitch:devolo.smartGroup.Group",
                "property.name": "value",
            },
        }
        spy = mocker.spy(self.homecontrol.updater, '_binary_switch')
        self.homecontrol.updater.update(message=message)
        spy.assert_not_called()

    def test__automatic_calibration(self):
        uid = self.devices['blinds']['uid']
        calibration_status = self.devices['blinds']['calibrationStatus']
        self.homecontrol.updater._automatic_calibration(
            message={"properties": {
                "uid": f"acs.{uid}",
                "property.value.new": {
                    "status": calibration_status,
                },
            }})
        assert self.homecontrol.devices[uid].settings_property['automatic_calibration'].calibration_status

    def test__automatic_calibration_key_error(self):
        uid = self.devices['blinds']['uid']
        calibration_status = self.devices['blinds']['calibrationStatus']
        self.homecontrol.updater._automatic_calibration(
            message={"properties": {
                "uid": f"acs.{uid}",
                "property.value.new": calibration_status,
            }})
        assert self.homecontrol.devices[uid].settings_property['automatic_calibration'].calibration_status

    def test__binary_async_blinds(self):
        uid = self.devices['blinds']['uid']
        self.homecontrol.devices[uid].settings_property['i2'].value = self.devices['blinds']['i2']
        self.homecontrol.updater._binary_async(
            message={"properties": {
                "uid": f"bas.{uid}#i2",
                "property.value.new": not self.devices['blinds']['i2'],
            }})
        assert not self.homecontrol.devices[uid].settings_property['i2'].value

    def test__binary_async_siren(self):
        uid = self.devices['siren']['uid']
        self.homecontrol.devices[uid].settings_property['muted'].value = self.devices['siren']['muted']
        self.homecontrol.updater._binary_async(
            message={"properties": {
                "uid": f"bas.{uid}",
                "property.value.new": not self.devices['siren']['muted'],
            }})
        assert not self.homecontrol.devices[uid].settings_property['muted'].value

    def test__binary_sensor(self):
        uid = self.devices['sensor']['uid']
        device = self.homecontrol.devices[uid].binary_sensor_property[f"devolo.BinarySensor:{uid}"]
        device.state = True
        state = device.state
        self.homecontrol.updater._binary_sensor(
            message={"properties": {
                "property.name": "state",
                "uid": f"devolo.BinarySensor:{uid}",
                "property.value.new": 0,
            }})
        state_new = device.state
        assert state != state_new
        assert device.last_activity != datetime.fromtimestamp(0)

    def test__binary_sync(self):
        uid = self.devices['blinds']['uid']
        self.homecontrol.updater._binary_sync(message={
            "properties": {
                "uid": f"bss.{uid}",
                "property.value.new": self.devices['blinds']['movement_direction'],
            },
        })
        assert self.homecontrol.devices[uid].settings_property["movement_direction"].direction is \
            bool(self.devices['blinds']['movement_direction'])

    def test__binary_switch(self):
        uid = self.devices.get("mains").get("uid")
        self.homecontrol.devices.get(uid).binary_switch_property.get(f"devolo.BinarySwitch:{uid}").state = True
        state = self.homecontrol.devices.get(uid).binary_switch_property.get(f"devolo.BinarySwitch:{uid}").state
        self.homecontrol.updater._binary_switch(message={
            "properties": {
                "property.name": "targetState",
                "uid": f"devolo.BinarySwitch:{uid}",
                "property.value.new": 0
            },
        })
        state_new = self.homecontrol.devices.get(uid).binary_switch_property.get(f"devolo.BinarySwitch:{uid}").state
        assert state != state_new

    def test__device_state(self):
        uid = self.devices.get("mains").get("uid")
        online_state = self.homecontrol.devices.get(uid).status
        self.homecontrol.updater._device_state(
            message={"properties": {
                "uid": uid,
                "property.name": "status",
                "property.value.new": 1
            }})
        assert self.homecontrol.devices.get(uid).status == 1
        assert online_state != self.homecontrol.devices.get(uid).status

    def test__general_device(self):
        uid = self.devices['mains']['uid']
        events_enabled = self.devices['mains']['properties']['eventsEnabled']
        self.homecontrol.devices[uid].settings_property['general_device_settings'].events_enabled = events_enabled
        self.homecontrol.updater._general_device(
            message={
                "properties": {
                    "uid": f"gds.{uid}",
                    "property.value.new": {
                        "eventsEnabled": not events_enabled,
                        "icon": self.devices['mains']['properties']['icon'],
                        "name": self.devices['mains']['properties']["itemName"],
                        "zoneID": self.devices['mains']['properties']["zoneId"],
                    },
                },
            })
        assert not self.homecontrol.devices[uid].settings_property['general_device_settings'].events_enabled

    def test__gateway_accessible(self):
        self.homecontrol.gateway.online = True
        self.homecontrol.gateway.sync = True
        accessible = self.homecontrol.gateway.online
        online_sync = self.homecontrol.gateway.sync
        self.homecontrol.updater._gateway_accessible(
            message={
                "properties": {
                    "property.name": "gatewayAccessible",
                    "property.value.new": {
                        "accessible": False,
                        "onlineSync": False,
                    },
                },
            })
        accessible_new = self.homecontrol.gateway.online
        online_sync_new = self.homecontrol.gateway.sync
        assert accessible != accessible_new
        assert online_sync != online_sync_new

    def test__grouping(self):
        zone_id = "hz_3"
        name = self.gateway['zones'][zone_id]
        self.homecontrol.updater._grouping(message={
            "properties": {
                "property.value.new": [{
                    "id": zone_id,
                    "name": self.devices['mains']['properties']['zone'],
                }],
            },
        })
        assert self.homecontrol.gateway.zones[zone_id] != name
        assert self.homecontrol.gateway.zones[zone_id] == self.devices['mains']['properties']['zone']

    def test__gui_enabled(self):
        uid = self.devices['mains']['uid']
        element_uids = self.devices['mains']['properties']['elementUIDs']
        enabled = self.devices['mains']['properties']['guiEnabled']
        self.homecontrol.devices[uid].binary_switch_property[element_uids[1]].enabled = enabled
        self.homecontrol.updater._gui_enabled(message={
            "uid": element_uids[0],
            "property.value.new": not enabled,
        })
        assert self.homecontrol.devices[uid].binary_switch_property[element_uids[1]].enabled != enabled

    def test__humidity_bar(self):
        uid = self.devices.get("humidity").get("uid")
        self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").value = 75
        self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").zone = 1
        current_value = self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").value
        current_zone = self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").zone
        self.homecontrol.updater._humidity_bar(
            message={"properties": {
                "uid": f"devolo.HumidityBarValue:{uid}",
                "property.value.new": 50,
            }})
        current_value_new = self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").value
        self.homecontrol.updater._humidity_bar(
            message={"properties": {
                "uid": f"devolo.HumidityBarZone:{uid}",
                "property.value.new": 0,
            }})
        current_zone_new = self.homecontrol.devices.get(uid).humidity_bar_property.get(f"devolo.HumidityBar:{uid}").zone

        assert current_value_new == 50
        assert current_zone_new == 0
        assert current_value != current_value_new
        assert current_zone != current_zone_new

    @pytest.mark.usefixtures("mock_publisher_add_event")
    @pytest.mark.usefixtures("mock_publisher_delete_event")
    @pytest.mark.parametrize("event", ["add", "delete"])
    def test__inspect_devices_valid(self, mocker, event):
        self.homecontrol.updater.on_device_change = lambda device_uids: (device_uids, event)
        spy = mocker.spy(self.homecontrol.updater._publisher, f"{event}_event")
        self.homecontrol.updater._inspect_devices(
            message={"properties": {
                "uid": "devolo.DevicesPage",
                "property.value.new": [],
            }})
        spy.assert_called_once()

    def test__inspect_devices_error(self, mocker):
        self.homecontrol.updater.on_device_change = None
        spy = mocker.spy(self.homecontrol.updater._logger, "error")
        self.homecontrol.updater._inspect_devices({})
        spy.assert_called_once_with("on_device_change is not set.")
        self.homecontrol.updater.on_device_change = lambda: None
        spy = mocker.spy(self.homecontrol.updater, "on_device_change")
        self.homecontrol.updater._inspect_devices(message={"properties": {
            "uid": "",
            "property.value.new": [],
        }})
        spy.assert_not_called()

    def test__meter(self):
        uid = self.devices.get("mains").get("uid")
        self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").current = 5
        self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").total = 230
        total = self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").total
        # Changing current value
        self.homecontrol.updater._meter(
            message={"properties": {
                "property.name": "currentValue",
                "uid": f"devolo.Meter:{uid}",
                "property.value.new": 7,
            }})
        current_new = self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").current
        # Check if current value has changed
        assert current_new == 7
        # Check if total has not changed
        assert total == self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").total
        # Changing total value
        self.homecontrol.updater._meter(
            message={"properties": {
                "property.name": "totalValue",
                "uid": f"devolo.Meter:{uid}",
                "property.value.new": 235
            }})
        total_new = self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").total
        # Check if total value has changed
        assert total_new == 235
        # Check if current value has not changed
        assert self.homecontrol.devices.get(uid).consumption_property.get(f"devolo.Meter:{uid}").current == current_new

    def test___multilevel_async(self):
        uid = self.devices['mains']['UID']
        element_uid = self.devices['mains']['properties']['settingUIDs'][4]
        value = self.devices['mains']['flashMode'] + 1
        self.homecontrol.updater._multilevel_async(message={"properties": {
            "uid": element_uid,
            "property.value.new": value
        }})
        assert self.homecontrol.devices[uid].settings_property['flash_mode'].value == value

    def test__multi_level_sensor(self):
        uid = self.devices['sensor']['uid']
        element_uid = f"devolo.MultiLevelSensor:{uid}#MultilevelSensor(1)"
        self.homecontrol.devices[uid].multi_level_sensor_property[element_uid].value = 100
        current = self.homecontrol.devices[uid].multi_level_sensor_property[element_uid].value
        self.homecontrol.updater._multi_level_sensor(message={"properties": {
            "uid": element_uid,
            "property.value.new": 50
        }})
        current_new = self.homecontrol.devices[uid].multi_level_sensor_property[element_uid].value

        assert current_new == 50
        assert current != current_new

    def test__multi_level_switch(self):
        device = self.devices.get("multi_level_switch")
        uid = device.get("uid")
        self.homecontrol.devices.get(uid).multi_level_switch_property.get(device.get("elementUIDs")[0]).value = \
            device.get("value")
        current = self.homecontrol.devices.get(uid).multi_level_switch_property.get(device.get("elementUIDs")[0]).value
        self.homecontrol.updater._multi_level_switch(
            message={"properties": {
                "uid": device.get("elementUIDs")[0],
                "property.value.new": device.get("max"),
            }})
        current_new = self.homecontrol.devices.get(uid).multi_level_switch_property.get(device.get("elementUIDs")[0]).value

        assert current_new == device.get("max")
        assert current != current_new

    def test__multilevel_sync_sensor(self):
        device = self.devices['sensor']
        uid = device['uid']
        value = device['properties']['value']
        self.homecontrol.devices[uid].settings_property['motion_sensitivity'].motion_sensitivity = value
        self.homecontrol.updater._multilevel_sync(
            message={"properties": {
                "uid": f"mss.{uid}",
                "property.value.new": value - 1,
            }})
        assert self.homecontrol.devices[uid].settings_property['motion_sensitivity'].motion_sensitivity == value - 1

    def test__multilevel_sync_shutter(self):
        device = self.devices['blinds']
        uid = device['uid']
        value = device['shutter_duration']
        self.homecontrol.devices[uid].settings_property['shutter_duration'].shutter_duration = value
        self.homecontrol.updater._multilevel_sync(
            message={"properties": {
                "uid": f"mss.{uid}",
                "property.value.new": value - 1,
            }})
        assert self.homecontrol.devices[uid].settings_property['shutter_duration'].shutter_duration == value - 1

    def test__multilevel_sync_siren(self):
        device = self.devices['siren']
        uid = device['uid']
        value = device['properties']['value']
        self.homecontrol.devices[uid].settings_property['tone'].tone = value
        self.homecontrol.updater._multilevel_sync(
            message={"properties": {
                "uid": f"mss.{uid}",
                "property.value.new": value - 1,
            }})
        assert self.homecontrol.devices[uid].settings_property['tone'].tone == value - 1

    def test__led(self):
        device = self.devices['mains']
        uid = device['uid']
        led_setting = device['properties']['led_setting']
        self.homecontrol.devices[uid].settings_property['led'].led_setting = led_setting
        self.homecontrol.updater._led(message={"properties": {
            "uid": f"lis.{uid}",
            "property.value.new": not led_setting,
        }})
        assert self.homecontrol.devices[uid].settings_property['led'].led_setting is not led_setting

    def test__parameter(self):
        device = self.devices['mains']
        uid = device['uid']
        param_changed = device['properties']['param_changed']
        self.homecontrol.devices[uid].settings_property['param_changed'].param_changed = param_changed
        self.homecontrol.updater._parameter(
            message={"properties": {
                "uid": f"cps.{uid}",
                "property.value.new": not param_changed,
            }})
        assert self.homecontrol.devices[uid].settings_property['param_changed'].param_changed is not param_changed

    def test__pending_operations_false(self):
        device = self.devices['mains']
        uid = device['uid']
        pending_operation = device['properties']['pending_operations']
        self.homecontrol.devices[uid].pending_operation = not pending_operation
        self.homecontrol.updater._pending_operations(message={"properties": {
            "uid": device['elementUIDs'][1]
        }})
        assert self.homecontrol.devices[uid].pending_operation

    def test__pending_operations_true(self):
        device = self.devices['mains']
        uid = device['uid']
        pending_operation = device['properties']['pending_operations']
        self.homecontrol.devices[uid].pending_operation = pending_operation
        self.homecontrol.updater._pending_operations(
            message={"properties": {
                "uid": device['elementUIDs'][1],
                "property.value.new": {
                    "status": 1
                }
            }})
        assert not self.homecontrol.devices[uid].pending_operation

    def test__pending_operations_setting(self):
        device = self.devices['mains']
        uid = device['uid']
        pending_operation = device['properties']['pending_operations']
        self.homecontrol.devices[uid].pending_operation = pending_operation
        self.homecontrol.updater._pending_operations(
            message={"properties": {
                "uid": device['settingUIDs'][3],
                "property.value.new": {
                    "status": 1,
                },
            }})
        assert not self.homecontrol.devices[uid].pending_operation

    def test__pending_operations_useless(self, mocker):
        spy = mocker.spy(self.homecontrol.updater._publisher, 'dispatch')
        self.homecontrol.updater._pending_operations(message={"properties": {
            "uid": "devolo.PairDevice",
        }})
        spy.assert_not_called()

    def test__protection_local(self):
        device = self.devices['mains']
        uid = device['uid']
        local_switch = device['properties']['local_switch']
        self.homecontrol.devices[uid].settings_property['protection'].local_switching = local_switch
        self.homecontrol.updater._protection(message={
            "properties": {
                "uid": f"ps.{uid}",
                "property.name": "targetLocalSwitch",
                "property.value.new": not local_switch,
            },
        })
        assert self.homecontrol.devices[uid].settings_property['protection'].local_switching is not local_switch

    def test__protection_remote(self):
        device = self.devices['mains']
        uid = device['uid']
        remote_switch = device['properties']['remote_switch']
        self.homecontrol.devices[uid].settings_property['protection'].remote_switching = remote_switch
        self.homecontrol.updater._protection(message={
            "properties": {
                "uid": f"ps.{uid}",
                "property.name": "targetRemoteSwitch",
                "property.value.new": not remote_switch,
            },
        })
        assert self.homecontrol.devices[uid].settings_property['protection'].remote_switching is not remote_switch

    def test__remote_control(self):
        device = self.devices.get("remote")
        uid = device.get("uid")
        self.homecontrol.devices.get(uid).remote_control_property \
            .get(device.get("elementUIDs")[0]).key_pressed = 0
        self.homecontrol.updater._remote_control(
            message={"properties": {
                "uid": device.get("elementUIDs")[0],
                "property.value.new": 1,
            }})

        assert self.homecontrol.devices.get(uid).remote_control_property \
            .get(device.get("elementUIDs")[0]).key_pressed == 1

    def test__since_time(self):
        device = self.devices['mains']
        uid = device['uid']
        now = datetime.now()
        total_since = self.homecontrol.devices[uid].consumption_property[f'devolo.Meter:{uid}'].total_since
        self.homecontrol.updater._since_time({
            "uid": f"devolo.Meter:{uid}",
            "property.value.new": now.replace(tzinfo=timezone.utc).timestamp() * 1000,
        })
        new_total_since = self.homecontrol.devices[uid].consumption_property[f'devolo.Meter:{uid}'].total_since
        assert total_since != new_total_since
        assert new_total_since == now

    def test__switch_type(self):
        device = self.devices['remote']
        uid = device['uid']
        self.homecontrol.updater._switch_type(
            message={"properties": {
                "uid": f"sts.{uid}",
                "property.value.new": device['key_count'] / 4,
            }})
        assert self.homecontrol.devices[uid].settings_property['switch_type'].value == device['key_count'] / 2

    def test__temperature_report(self):
        device = self.devices['sensor']
        uid = device['uid']
        self.homecontrol.devices[uid].settings_property['temperature_report'].temp_report = device['temp_report']
        self.homecontrol.updater._temperature_report(
            message={"properties": {
                "uid": f"trs.{uid}",
                "property.value.new": not device['temp_report'],
            }})
        assert self.homecontrol.devices[uid].settings_property['temperature_report'].temp_report is not device['temp_report']

    def test__voltage_multi_level_sensor(self):
        uid = self.devices.get("mains").get("uid")
        self.homecontrol.devices.get(uid).multi_level_sensor_property.get(f"devolo.VoltageMultiLevelSensor:{uid}").value = 231
        current = self.homecontrol.devices.get(uid).multi_level_sensor_property.get(
            f"devolo.VoltageMultiLevelSensor:{uid}").value
        self.homecontrol.updater._voltage_multi_level_sensor(
            message={"properties": {
                "uid": f"devolo.VoltageMultiLevelSensor:{uid}",
                "property.value.new": 234
            }})
        current_new = self.homecontrol.devices.get(uid).multi_level_sensor_property.get(
            f"devolo.VoltageMultiLevelSensor:{uid}").value

        assert current_new == 234
        assert current != current_new
